#!/usr/bin/python

#  fdxbnum.py - generate / decode FDX-B EM4x05 compliant IDs
# 
#  Adam Laurie <adam@algroup.co.uk>
#  http://rfidiot.org/
# 
#  This code is copyright (c) Adam Laurie, 2006, All rights reserved.
#  For non-commercial use only, the following terms apply - for all other
#  uses, please contact the author:
#
#    This code is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#


import RFIDIOt
import RFIDIOtconfig
import sys
import os
import string

card= RFIDIOtconfig.card
card.info('fdxbnum v0.1a')

precoded= False

if len(sys.argv) == 2 or len(sys.argv) == 3:
	print "Decode: "
	if len(sys.argv[1]) == 16:
        	card.FDXBIDPrint(sys.argv[1])
	else:
		card.FDXBIDPrint(sys.argv[1][1:])
	if len(sys.argv) == 3:
		if sys.argv[2] == 'WRITE':
			precoded= True
		else:
			print 'Unrecognised option: ' + sys.argv[2]
			os._exit(False)
	else:
		os._exit(True)

if len(sys.argv) >= 4 or precoded:
	if precoded:
		id= sys.argv[1]
	else:
		print "Encode: ",
		id= card.FDXBIDEncode(sys.argv[1],sys.argv[2],sys.argv[3])
		print id
	# bit 64 will be transmitted first, so leave string reversed
	idbin= card.ToBinaryString(card.ToBinary(id))
	# construct FDX-B encoded blocks
	out= ''
	# header is ten zeros and a '1'
	header= '00000000001'
	out += header
	# break id into 8 bit chunks with a trailing '1' on each
	for x in range(0,len(idbin),8):
		out += idbin[x:x+8] + '1'
	# add 16 CRC-CCITT error detection bits
	crc= '%04x' % card.crcccitt(card.ToBinary(id))
	crcbin= card.ToBinaryString(card.ToBinary(crc))
	# crc is transmitted LSB first with traling '1's
	out += crcbin[0:8] + '1'
	out += crcbin[8:16] + '1'
	# add 3 sets of trailer bits (RFU)
	trailer= '000000001'
	for x in range(3):
		out += trailer
	if (len(sys.argv) == 5 and sys.argv[4] == 'WRITE') or precoded:
       		while True:
			# Q5 must be forced into Q5 mode to be sure of detection so try that first 
			card.settagtype(card.Q5)
			card.select()
			if not card.tagtype == card.Q5:
				card.settagtype(card.ALL)
               		print
               		print '  Card ID: ' + card.select()
			if card.tagtype == card.Q5 or card.tagtype == card.HITAG2:
               			x= string.upper(raw_input('  *** Warning! This will overwrite TAG! Proceed (y/n)? '))
               			if x == 'N':
                       			os._exit(False)
               			if x == 'Y':
                       			break
			else:
				x= raw_input('  Incompatible TAG! Hit <RETURN> to retry...')
		writetag= True
		print
	else:
		writetag= False
	# now turn it all back to 4 byte hex blocks for writing
	outbin= ''
	outhex= ['','','','','']
	# control block for Q5:
	# carrier 32 (2 * 15 + 2)
	# rf/? (don't care) - set to 00
	# data inverted
	# biphase
	# maxblock 4
	print '  Q5 Control Block:  ',
	q5control= '6000F0E8'
	print q5control
	for x in range(0,len(out),8):
		outbin += chr(int(out[x:x + 8],2))
	for x in range(0,len(outbin),4):
		print '    Q5 Data Block %02d:' % (x / 4 + 1),
		outhex[x / 4 + 1]= card.ToHex(outbin[x:x+4])
		print outhex[x / 4 + 1]
	# control block for Hitag2
	# Public Mode B
	# default password
	print
	print '  Hitag2 Control Block:  ',
	h2control= '00' + card.HITAG2_TRANSPORT_TAG
	print h2control
	for x in range(1,5,1):
		print '    Hitag2 Data Block %02d:' % (x + 3),
		print outhex[x]
	if writetag == True:
		print 
		print '  Writing to tag type: ' + card.LFXTags[card.tagtype]
		if card.tagtype == card.Q5:
			outhex[0]= q5control
			offset= 0
		if card.tagtype == card.HITAG2:
			outhex[0]= h2control
			offset= 3
			card.login('','',card.HITAG2_TRANSPORT_RWD)
		for x in range(4 + offset,-1 + offset,-1):
        		if not card.writeblock(x,outhex[x - offset]):
				# we expect a Q5 to fail after writing the control block as it re-reads
				# it before trying to verify the write and switches mode so is now no longer in Q5 mode
				if x == offset:
					print '    Control:  ' + outhex[x - offset]
					print
					print '  Done!'
					# now check for FDX-B ID
               				card.settagtype(card.EM4x05)
					print '  Card ID: ' + card.select()
				else:
                			print 'Write failed!'
                			os._exit(False)
			else:
				# hitag2 don't change mode until the next time they're selected so write
				# confirmation of control block should be ok
				if x == offset:
					print '    Control:  ' + outhex[x - offset]
					print
					print '  Done!'
					# now check for FDX-B ID
               				card.settagtype(card.EM4x05)
					print '  Card ID: ' + card.select()
				else:
					print '    Block %02d: %s' % (x , outhex[x - offset])
               	card.settagtype(card.ALL)
	os._exit(True)
print sys.argv[0] + ' - generate / decode FDX-B EM4x05 compliant IDs'
print 'Usage: ' + sys.argv[0] + ' <ID> [WRITE] | <APPID> <COUNTRY CODE> <NATIONAL ID> [WRITE]'
print
print '\tIf a single 16 HEX digit ID is provided, it will be decoded according to the FDX-B standard.'
print '\tAlternatively, specifying a 4 HEX digit Application ID, 3 or 4 digit decimal country code'
print '\t(normally based on ISO-3166 country codes or ICAR.ORG manufacturer codes, range 0 - 4095)'
print '\tand a decimal National ID Number will generate a 16 HEX digit ID.'
print '\tNote: Application ID 8000 is \'Animal\', and 0000 is non-Animal.'
print
print '\tIf the WRITE option is specified, a Q5 or Hitag2 will be prgrammed to emulate FDX-B.'
print '\tNote that while Q5 tags can be reverted to Q5 mode with the \'q5unfuck.py\' program,'
print '\tHitag2 tags cannot currently be recovered.'
print
