#!/usr/bin/python

#  unique.py - 	generate EM4x02 and/or UNIQUE compliant IDs
#  		these can then be written to a Q5 tag to emulate EM4x02
#		by transmitting data blocks 1 & 2 (MAXBLOCK == 2).
# 
#  Adam Laurie <adam@algroup.co.uk>
#  http://rfidiot.org/
# 
#  This code is copyright (c) Adam Laurie, 2006, All rights reserved.
#  For non-commercial use only, the following terms apply - for all other
#  uses, please contact the author:
#
#    This code is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#


import RFIDIOt
import RFIDIOtconfig
import sys
import os
import string

card= RFIDIOtconfig.card
card.info('readlfx v0.1c')

# standard config block
CFB='e601f004'

def uniquetoem(data):
	# swap words
	tmp= ''
	for x in range(5):
		tmp += data[x * 2 + 1] + data[x * 2]
	# reverse bits
	return card.ToBinaryString(card.ToBinary(tmp))[::-1]

if len(sys.argv) < 3 or len(sys.argv) > 4:
	print sys.argv[0] + ' - generate EM4x02 and/or UNIQUE compliant ID data blocks'
	print 'Usage: ' + sys.argv[0] + ' <TYPE> <ID> [WRITE]'
	print
	print '\t10 digit HEX ID will be translated to valid data for blocks 1 & 2'
	print '\tfor a Q5 tag running in EM4x02 emulation mode, where TYPE is U for'
	print '\tUNIQUE code and E for EM4x02. For guidance, a standard emulation'
	print '\tcontrol block (0) will also be displayed.' 
	print 
	print '\tIf the optional WRITE argument is specified, programming a Q5 tag will '
	print '\tbe initiated.'
	print
	os._exit(True)

id= sys.argv[2]
if sys.argv[1] == 'E':
	type= 'EM4x02'
	idbin= uniquetoem(id)
else:
	if sys.argv[1] == 'U':
		type= 'UNIQUE'
		idbin= card.ToBinaryString(card.ToBinary(id))
	else:
		print 'Unknown TYPE: ' + sys.argv[1]
		os._exit(False)
# standard header == 9 bits of '1'
out= '111111111'
# break output into 4 bit chunks and add parity
colparity= [0,0,0,0]
for x in range(0,len(idbin),4):
	parity= 0
	chunk= idbin[x:x+4]
	for y in range(4):
		parity += int(chunk[y],2)
		colparity[y] += int(chunk[y],2)
	out += chunk + '%s' % (int(parity) % 2)
# add column parity
for x in range(4):
	out += '%s' % (int(colparity[x]) % 2)
# add stop bit
out += '0'
db1= '%08x' % int(out[:32],2)
db2= '%08x' % int(out[32:64],2)
print
print '  ' + type + ' ID: ' + id
print '  Q5 ID: ' + '%08x' % int(idbin[:32],2)
if type ==  'EM4x02':
	print '  UNIQUE ID: ' + '%10x' % int(idbin,2)
else:
	print '  EM4x02 ID: ' + ('%10x' % int(uniquetoem(id),2))[::-1]
print '  Binary traceablility data: ' + out
print
print '  Standard Config Block (0): ' + CFB
print '  Data Block 1: ' + db1
print '  Data Block 2: ' + db2

if len(sys.argv) == 4 and sys.argv[3] == 'WRITE':
	card.settagtype(card.Q5)
	while True:
		print
		print '  Card ID: ' + card.select()
        	x= string.upper(raw_input('  *** Warning! This will overwrite TAG! Proceed (y/n)? '))
        	if x == 'N':
                	os._exit(False)
        	if x == 'Y':
			break
	print 'Writing...'
	if not card.writeblock(0,CFB) or not card.writeblock(1,db1) or not card.writeblock(2,db2):
		print 'Write failed!'
		os._exit(False)
	else:
		card.settagtype(card.ALL)
		print 'Done!'
